"""Middleware for InvenTree."""

from django.http import HttpResponse
from django.urls import resolve


class AuthRequiredMiddleware(object):
    """Check for user to be authenticated."""

    def __init__(self, get_response):
        """Save response object."""
        self.get_response = get_response

    def __call__(self, request):
        """Check if user needs to be authenticated and is.

        Redirects to login if not authenticated.
        """
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        assert hasattr(request, "user")

        # Is the function exempt from auth requirements?
        path_func = resolve(request.path).func
        if getattr(path_func, "auth_exempt", False) is True:
            return self.get_response(request)

        if not request.user.is_authenticated:
            # No authorization was found for the request
            # Return a 401 (Unauthorized) response code for this request
            return HttpResponse("Unauthorized", status=401)

        response = self.get_response(request)

        return response



from typing import Any

from django.conf import LazySettings

DEFAULT_SETTINGS = {
    "EXTRA_URL_SCHEMES": [],
}


class NoValue:
    """A custom class that shows nothing was found."""


class CustomLazySettings(LazySettings):
    """Custom settings class to replace django settings.

    This class is used to provide defaults for settings which are not defined in the configuration file.
    These are placed on top of the default settings provided by django.

    Callflow:
    1. django.conf.settings
    2. accu.core.settings.DEFAULT_SETTINGS
    """

    def __getattr__(self, item: Any) -> Any:
        """Override the __getattr__ method to provide default values for settings.

        Args:
            item: The setting name

        Returns:
            The setting value
        """

        try:
            # Lookup the setting the django way
            result = super().__getattr__(item)
        except AttributeError:
            # Not Found, try to get the default value
            result = DEFAULT_SETTINGS.get(item, NoValue)
            if result is NoValue:
                raise AttributeError(f"Setting '{item}' is not defined")

        return result


settings = CustomLazySettings()



"""Helper functions for loading InvenTree configuration options."""

import os
import random
import shutil
import string
from pathlib import Path

import yaml

from .helpers.logging import LOGGER
from .utils import get_base_dir


def get_config_file(create=True) -> Path:
    """Returns the path of the InvenTree configuration file.

    Note: It will be created it if does not already exist!
    """
    base_dir = get_base_dir()

    cfg_filename = os.getenv("INVENTREE_CONFIG_FILE")

    if cfg_filename:
        cfg_filename = Path(cfg_filename.strip()).resolve()
    else:
        # Config file is *not* specified - use the default
        cfg_filename = base_dir.joinpath("config.yaml").resolve()

    if not cfg_filename.exists() and create:
        print(
            "InvenTree configuration file 'config.yaml' not found - creating default file"
        )

        cfg_template = base_dir.joinpath("config_template.yaml")
        shutil.copyfile(cfg_template, cfg_filename)
        print(f"Created config file {cfg_filename}")

    return cfg_filename


def load_config_data() -> map:
    """Load configuration data from the config file."""

    cfg_file = get_config_file()

    with open(cfg_file, "r") as cfg:
        data = yaml.safe_load(cfg)

    return data


def get_setting(env_var=None, config_key=None, default_value=None, typecast=None):
    """Helper function for retrieving a configuration setting value.

    - First preference is to look for the environment variable
    - Second preference is to look for the value of the settings file
    - Third preference is the default value

    Arguments:
        env_var: Name of the environment variable e.g. 'INVENTREE_STATIC_ROOT'
        config_key: Key to lookup in the configuration file
        default_value: Value to return if first two options are not provided
        typecast: Function to use for typecasting the value
    """

    def try_typecasting(value):
        """Attempt to typecast the value"""
        if typecast is not None:
            # Try to typecast the value
            try:
                return typecast(value)
            except Exception as error:
                LOGGER.error(
                    f"Failed to typecast '{env_var}' with value '{value}' to type '{typecast}' with error {error}"
                )
        return value

    # First, try to load from the environment variables
    if env_var is not None:
        val = os.getenv(env_var, None)

        if val is not None:
            return try_typecasting(val)

    # Next, try to load from configuration file
    if config_key is not None:
        cfg_data = load_config_data()

        result = None

        # Hack to allow 'path traversal' in configuration file
        for key in config_key.strip().split("."):

            if type(cfg_data) is not dict or key not in cfg_data:
                result = None
                break

            result = cfg_data[key]
            cfg_data = cfg_data[key]

        if result is not None:
            return try_typecasting(result)

    # Finally, return the default value
    return try_typecasting(default_value)


def get_settings_dir(reference: str, create=True):
    """Return the absolute path for a directory."""

    dir = get_setting(f"INVENTREE_{reference.upper()}", reference)

    if not dir:
        raise FileNotFoundError(f"{reference} not specified")

    dir_path = Path(dir).resolve()

    if create:
        dir_path.mkdir(parents=True, exist_ok=True)

    return dir_path


def get_plugin_file():
    """Returns the path of the InvenTree plugins specification file.

    Note: It will be created if it does not already exist!
    """

    # Check if the plugin.txt file (specifying required plugins) is specified
    plugin_file = get_setting("INVENTREE_PLUGIN_FILE", "plugin_file")

    if not plugin_file:
        # If not specified, look in the same directory as the configuration file
        config_dir = get_config_file().parent
        plugin_file = config_dir.joinpath("plugins.txt")
    else:
        # Make sure we are using a modern Path object
        plugin_file = Path(plugin_file)

    if not plugin_file.exists():
        LOGGER.warning(
            "Plugin configuration file does not exist - creating default file"
        )
        LOGGER.info(f"Creating plugin file at '{plugin_file}'")

        # If opening the file fails (no write permission, for example), then this will throw an error
        plugin_file.write_text(
            "# InvenTree Plugins (uses PIP framework to install)\n\n"
        )

    return plugin_file


def get_secret_key():
    """Return the secret key value which will be used by django.

    Following options are tested, in descending order of preference:

    A) Check for environment variable INVENTREE_SECRET_KEY => Use raw key data
    B) Check for environment variable INVENTREE_SECRET_KEY_FILE => Load key data from file
    C) Look for default key file "secret_key.txt"
    D) Create "secret_key.txt" if it does not exist
    """

    # Look for environment variable
    if secret_key := get_setting("INVENTREE_SECRET_KEY", "secret_key"):
        LOGGER.info("SECRET_KEY loaded by INVENTREE_SECRET_KEY")  # pragma: no cover
        return secret_key

    # Look for secret key file
    if secret_key_file := get_setting("INVENTREE_SECRET_KEY_FILE", "secret_key_file"):
        secret_key_file = Path(secret_key_file).resolve()
    else:
        # Default location for secret key file
        secret_key_file = get_base_dir().joinpath("secret_key.txt").resolve()

    if not secret_key_file.exists():
        LOGGER.info(f"Generating random key file at '{secret_key_file}'")

        # Create a random key file
        options = string.digits + string.ascii_letters + string.punctuation
        key = "".join([random.choice(options) for i in range(100)])
        secret_key_file.write_text(key)

    LOGGER.info(f"Loading SECRET_KEY from '{secret_key_file}'")

    key_data = secret_key_file.read_text().strip()

    return key_data


def get_custom_file(
    env_ref: str, conf_ref: str, log_ref: str, lookup_media: bool = False
):
    """Returns the checked path to a custom file.

    Set lookup_media to True to also search in the media folder.
    """
    from django.contrib.staticfiles.storage import StaticFilesStorage
    from django.core.files.storage import default_storage

    value = get_setting(env_ref, conf_ref, None)

    if not value:
        return None

    static_storage = StaticFilesStorage()

    if static_storage.exists(value):
        LOGGER.info(f"Loading {log_ref} from static directory: {value}")
    elif lookup_media and default_storage.exists(value):
        LOGGER.info(f"Loading {log_ref} from media directory: {value}")
    else:
        add_dir_str = " or media" if lookup_media else ""
        LOGGER.warning(
            f"The {log_ref} file '{value}' could not be found in the static{add_dir_str} directories"
        )
        value = False

    return value



"""Functions for tasks and a few general async tasks."""

import warnings

from django.conf import settings
from django.core.exceptions import AppRegistryNotReady
from django.db.utils import OperationalError, ProgrammingError

from ..helpers.logging import LOGGER


def schedule_task(taskname, **kwargs):
    """Create a scheduled task.

    If the task has already been scheduled, ignore!
    """
    # If unspecified, repeat indefinitely
    repeats = kwargs.pop("repeats", -1)
    kwargs["repeats"] = repeats

    try:
        from django_q.models import Schedule
    except AppRegistryNotReady:  # pragma: no cover
        LOGGER.info("Could not start background tasks - App registry not ready")
        return

    try:
        # If this task is already scheduled, don't schedule it again
        # Instead, update the scheduling parameters
        if Schedule.objects.filter(func=taskname).exists():
            LOGGER.debug(f"Scheduled task '{taskname}' already exists - updating!")

            Schedule.objects.filter(func=taskname).update(**kwargs)
        else:
            LOGGER.info(f"Creating scheduled task '{taskname}'")

            Schedule.objects.create(name=taskname, func=taskname, **kwargs)
    except (OperationalError, ProgrammingError):  # pragma: no cover
        # Required if the DB is not ready yet
        pass


def raise_warning(msg):
    """Log and raise a warning."""
    LOGGER.warning(msg)

    # If testing is running raise a warning that can be asserted
    if settings.TESTING:
        warnings.warn(msg)


def offload_task(taskname, *args, force_async=False, force_sync=False, **kwargs):
    """Create an AsyncTask if workers are running. This is different to a 'scheduled' task, in that it only runs once!

    If workers are not running or force_sync flag
    is set then the task is ran synchronously.
    """
    try:
        import importlib

        from django_q.tasks import AsyncTask

        from InvenTree.status import is_worker_running
    except AppRegistryNotReady:  # pragma: no cover
        LOGGER.warning(f"Could not offload task '{taskname}' - app registry not ready")
        return
    except (OperationalError, ProgrammingError):  # pragma: no cover
        raise_warning(f"Could not offload task '{taskname}' - database not ready")

    if force_async or (is_worker_running() and not force_sync):
        # Running as asynchronous task
        try:
            task = AsyncTask(taskname, *args, **kwargs)
            task.run()
        except ImportError:
            raise_warning(f"WARNING: '{taskname}' not started - Function not found")
    else:

        if callable(taskname):
            # function was passed - use that
            _func = taskname
        else:
            # Split path
            try:
                app, mod, func = taskname.split(".")
                app_mod = app + "." + mod
            except ValueError:
                raise_warning(
                    f"WARNING: '{taskname}' not started - Malformed function path"
                )
                return

            # Import module from app
            try:
                _mod = importlib.import_module(app_mod)
            except ModuleNotFoundError:
                raise_warning(
                    f"WARNING: '{taskname}' not started - No module named '{app_mod}'"
                )
                return

            # Retrieve function
            try:
                _func = getattr(_mod, func)
            except AttributeError:  # pragma: no cover
                # getattr does not work for local import
                _func = None

            try:
                if not _func:
                    _func = eval(func)  # pragma: no cover
            except NameError:
                raise_warning(
                    f"WARNING: '{taskname}' not started - No function named '{func}'"
                )
                return

        # Workers are not running: run it as synchronous task
        _func(*args, **kwargs)



from .settings import settings
from .validators import allowable_url_schemes

# Initialize the settings
settings.configure()


def test_true():
    assert True


def test_allowable_url_schemes():
    """Test the allowable_url_schemes function."""
    assert allowable_url_schemes() == ["http", "https", "ftp", "ftps", "sftp"]

    settings.EXTRA_URL_SCHEMES = ["weddav"]
    assert "weddav" in allowable_url_schemes()



"""Custom field validators for InvenTree."""

from .settings import settings


def allowable_url_schemes():
    """Return the list of allowable URL schemes.

    In addition to the default schemes allowed by Django,
    the install configuration file (config.yaml) can specify
    extra schemas
    """
    # Default schemes
    schemes = ["http", "https", "ftp", "ftps", "sftp"]

    extra = settings.EXTRA_URL_SCHEMES

    for e in extra:
        if e.lower() not in schemes:
            schemes.append(e.lower())

    return schemes



from unittest import TestCase

from .config import get_plugin_file


def test_true():
    assert True


class TestConfig(TestCase):
    def test_get_plugin_file(self):
        """Test that the plugin file is found.

        # TODO This test is not working, because the plugin template file is not found.
        """
        with self.assertRaises(FileNotFoundError):
            get_plugin_file()



from .utils import is_true


def test_true():
    assert True


def test_is_true():
    """Test the is_true function."""
    # False assertations
    assert is_true("False") is False
    assert is_true("F") is False
    assert is_true("0") is False

    # True assertations
    assert is_true("1")
    assert is_true("y")
    assert is_true("yes")
    assert is_true("t")
    assert is_true("true")
    assert is_true("on")
